/* ---------------------------------------------------------------
 * ARCHIVO: mpi_pi_reduce.c
 * DESCRIPCION:  Programa MPI para el clculo del nmero  pi.
 *   Algoritmo utilizado: "dartboard".
 *   Esta versin de programa utiliza mpc_reduce para colectar resultados.
 * EVENTO: Congreso Nacional de Software Libre 2004(CONSOL) www.consol.org.mx/2004/
 * PONENTES:
 *    Castillo Castillo Jos - www.josecc.net
 *    Ricardo Benjamn Olvera Acosta
 *    Jeanette Moreno Snchez.
 * URL: http://www.columbia.edu/itc/applied/e6901/
 * NOTA: Slo se utiliza este cdigo como ejemplo de demostracin. Se ha modificado
 *    algunas lneas con fines didacticos, puede encontrar el cdigo original en
 *    el sitio antes referido.

 * --------------------------------------------------------------- */

#include "mpi.h"
#include <stdlib.h>
#include <stdio.h>

void srandom (unsigned seed);
double dboard (int darts);
long random(void);

#define DARTS 5001     /* number of throws at dartboard */
#define ROUNDS 10       /* number of times "darts" is iterated */
#define MASTER 0        /* task ID of master task */
#define sqr(x)	((x)*(x))

int main(argc,argv)
     int argc;
     char *argv[];
{
  double homepi,        /* value of pi calculated by current task */
    pisum,	        /* sum of tasks' pi values */
    pi,	        	/* average of pi after "darts" is thrown */
    avepi;	        /* average pi value for all iterations */

  int taskid,	        /* task ID - also used as seed number */
    numtasks,       	/* number of tasks */
    rc,             	/* return code */
    i;

  MPI_Status status;

  /* Obtain number of tasks and task ID */
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&numtasks);
  MPI_Comm_rank(MPI_COMM_WORLD,&taskid);
  printf ("MPI - proceso %d iniciado...\n", taskid);
  if (taskid == MASTER) 
    printf ("Usando %d procesos para calcular pi (3.1415926535)\n",numtasks);

  /* Set seed for random number generator equal to task ID */
  srandom (taskid);

  avepi = 0;
  for (i = 0; i < ROUNDS; i++)
    {
      /* All tasks calculate pi using dartboard algorithm */
      homepi = dboard(DARTS);

      /* Use MPI_Reduce to sum values of homepi across all tasks 
       * Master will store the accumulated value in pisum 
       * - homepi is the send buffer
       * - pisum is the receive buffer (used by the receiving task only)
       * - the size of the message is sizeof(double)
       * - MASTER is the task that will receive the result of the reduction
       *   operation
       * - MPI_SUM is a pre-defined reduction function (double-precision
       *   floating-point vector addition).  Must be declared extern.
       * - MPI_COMM_WORLD is the group of tasks that will participate.
       */

      rc = MPI_Reduce(&homepi, &pisum, 1, MPI_DOUBLE, MPI_SUM,
		      MASTER, MPI_COMM_WORLD);
      if (rc != MPI_SUCCESS)
	printf("%d: failure on mpc_reduce\n", taskid);

      /* Master computes average for this iteration and all iterations */
      if (taskid == MASTER)
	{
	  pi = pisum/(double)numtasks;
	  avepi = ((avepi * i) + pi)/(double)(i + 1); 
	  printf("   Despues de %8d iteraciones, el valor aproximado de pi = %10.8f\n",
		 (DARTS * (i + 1)),avepi);
	}    
    } 
  MPI_Finalize();
  return 0;
}

double dboard(int darts)
{
  double x_coord,       /* x coordinate, between -1 and 1  */
    y_coord,       /* y coordinate, between -1 and 1  */
    pi,            /* pi  */
    r;             /* random number between 0 and 1  */
  int score,            /* number of darts that hit circle */
    n;
  unsigned long cconst; /* used to convert integer random number */
  /* between 0 and 2^31 to double random number */
  /* between 0 and 1  */
                                                                                                                            
  cconst = 2 << (31 - 1);
  score = 0;
                                                                                                                            
  /* "throw darts at board" */
  for (n = 1; n <= darts; n++)  {
    /* generate random numbers for x and y coordinates */
    r = (double)random()/cconst;
    x_coord = (2.0 * r) - 1.0;
    r = (double)random()/cconst;
    y_coord = (2.0 * r) - 1.0;
                                                                                                                            
    /* if dart lands in circle, increment score */
    if ((sqr(x_coord) + sqr(y_coord)) <= 1.0)
      score++;
  }
                                                                                                                            
  /* calculate pi */
  pi = (double)4.0 * (double)score/(double)darts;
  printf("%f\n",pi);
  return(pi);
}

