#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <qimage.h>
#include <qtextstream.h>

#include <sys/types.h>
#include <signal.h>


#include "utility.h"
#include "task.h"
#include "job.h"


// --------------------------------------------------------------------------------
JOB::JOB(QString &error,int task,int job,const QString &cmdString,const QString &inputFile,const QString &attrString,const QString &outputDir,int width,int height,int start,int end) :
  m_cmdString(cmdString),m_attrString(attrString),m_outputDir(outputDir),m_inputFile(inputFile),m_pid(getpid()),m_width(width),m_height(height),m_start(start),m_end(end),m_task(task),m_job(job),m_status(QUEUED),m_exitStatus(0),m_rendering_p(0),m_frameNumber(-1),m_frameTotal(-1),m_item(0)
{
  char *tmp = tempnam(outputDir,"tmp");
  if(!tmp)
  {
    error = "Cannot create temp. file ! \nPlease check the configuration.";
    return;
  }

  m_outputFile = QString(tmp) + ".ppm" ;

  free(tmp);
}

// --------------------------------------------------------------------------------
JOB::~JOB(void)
{
  if(!m_outputFile.isEmpty())
  {
    if(m_frameNumber!=-1) unlink(getRealOutputName());
                     else unlink(m_outputFile);
  }
}

// --------------------------------------------------------------------------------
bool JOB::run(void)
{
  // Rendering line
  m_rendering_p = 0;
  
  // Clear exit status
  m_exitStatus = 0;
  
  // Clear streams
  m_stdOut = "";
  m_stdErr = "";

  // Clear arguments
  clearArguments();

  // Add arguments
  addArgument(m_cmdString);

  QString attrString(m_attrString);
  QTextIStream strStream(&attrString);
  
  while(!strStream.eof())
  { 
    QString str;
    
    strStream>>str;
    addArgument(str);

  }
    
  addArgument("-I"  + m_inputFile);
  addArgument("-O"  + m_outputFile);
  addArgument("-W"  + STRING(m_width));
  addArgument("-H"  + STRING(m_height));
  addArgument("-SR" + STRING(m_start));
  addArgument("-ER" + STRING(m_end+1));
  addArgument("+FP");
  addArgument("-D");

  connect(this,SIGNAL(processExited()),  this,SLOT(exitCallback()));
  connect(this,SIGNAL(readyReadStdout()),this,SLOT(stdOutCallback()));
  connect(this,SIGNAL(readyReadStderr()),this,SLOT(stdErrCallback()));

  if(!start())
  {
    return(false);
  }

  m_status = RUNNING;
  if(m_item) m_item->setText(4,"RUNNING");
  
  return(true);
}

// --------------------------------------------------------------------------------
void JOB::exitCallback(void)
{

  m_status = FINISHED;
  
  if(m_item) m_item->setText(4,"FINISHED");

  // Set the exit status
  m_exitStatus = exitStatus();
  
  if(m_exitStatus)
  {
    if(m_item)
    {
      QImage img = QImage::fromMimeSource("error.png");
      m_item->setPixmap(0,img);
    }

    m_rendering_p = 0;
  } else {
           if(m_item)
           {
             m_item->setPixmap(0,QImage());  
             m_item->setText(5,"100 %");
           }

            m_rendering_p = 100;
         }
  
  // Task callback
  TASK *t = TASK::findTask(m_task);
  if(!t)
  {
     CERR("Warning task not found !");
     return;
  }
  
  t->jobFinishCallback();
}

// --------------------------------------------------------------------------------
void JOB::stdOutCallback(void)
{
  QByteArray  a(readStdout());  
  QString     str;
  
  for (int i=0;i<(int)a.size();i++) 
    str += a[i];

  m_stdOut += str;
}

// --------------------------------------------------------------------------------
void JOB::stdErrCallback(void)
{
  QByteArray a(readStderr());  
  QString    str;
  
  for (int i=0;i<(int)a.size();i++) 
    str += a[i];

  m_stdErr += str;
  
  if(str.isEmpty()) return;

  char *p = strstr(str,"Rendering line");
  if(p)
  {
     QString      tmp(p);
     QTextStream  s(&tmp,IO_ReadOnly);
     QString      s1;
     int          line;
     
     s >> s1;
     s >> s1;
     s >> line;
     
     if(line>=0 && line<=m_end)
     {
       m_rendering_p = (int) (float) (100.0/(m_end-m_start+1) * line);
       
       if(m_item) m_item->setText(5,QString(STRING(m_rendering_p) + " %"));
     }
  }
}

// --------------------------------------------------------------------------------
const QString JOB::getJobStatusStr(void) const
{
  switch(m_status)
  {
    case QUEUED:
      return("QUEUED");

    case RUNNING:
      return("RUNNING");

    case FINISHED:
      return("FINISHED");

    case PAUSED:
      return("PAUSED");
            
    default:
    return("UNKNOWN");
  }
}

// --------------------------------------------------------------------------------
void JOB::stop(void)
{
  if(m_status==PAUSED) continue_job();
  
  tryTerminate();
  //kill();
}

// --------------------------------------------------------------------------------
void JOB::pause_job(void)
{
  int pid = getPID();
  
  int ret = ::kill(pid,SIGSTOP);
  
  if(ret!=0)
  {
    CERR("Warning : kill failed !");  
  }

  m_status = PAUSED;
}  

// --------------------------------------------------------------------------------
void JOB::continue_job(void)
{
  int pid = getPID();

  int ret = ::kill(pid,SIGCONT);

  if(ret!=0)
  {
    CERR("Warning : kill failed !");  
  }
    
  m_status = RUNNING;  
}

// --------------------------------------------------------------------------------
QString  JOB::getRealOutputName(void) const
{
  int field = strlen(STRING(getFrameTotal()));
  
  QString str1 = getOutputFile().mid(0,getOutputFile().length()-4);
  QString str2;
  str2.setNum(getFrameNumber());
  
  while((int) str2.length()!=field)
  {
    str2 = "0" + str2;
  }
  
  QString ret = str1  + str2 + ".ppm";
  
  return(ret);
}
